require 'assert'
require 'mr/after_commit/record'

require 'ardb/record_spy'
require 'much-plugin'
require 'mr/record'

module MR::AfterCommit::Record

  class UnitTests < Assert::Context
    desc "MR::AfterCommit::Record"
    subject{ MR::AfterCommit::Record }

    should "use much-plugin" do
      assert_includes MuchPlugin, subject
    end

  end

  class MixinTests < UnitTests
    desc "when mixed in"
    setup do
      @record_class = Ardb::RecordSpy.new do
        include MR::AfterCommit::Record
      end
    end
    subject{ @record_class }

    should "be an MR record" do
      assert_includes MR::Record, subject
    end

    should "use the records procs methods mixin" do
      assert_includes MR::AfterCommit::RecordProcsMethods, subject
    end

    should "add an after commit callback on create" do
      callback = subject.callbacks.find do |c|
        c.type == :after_commit && c.options[:on] == :create
      end

      assert_not_nil callback
      assert_equal :mr_after_commit_call_procs_for_create, callback.args.first
    end

    should "add an after commit callback on update" do
      callback = subject.callbacks.find do |c|
        c.type == :after_commit && c.options[:on] == :update
      end

      assert_not_nil callback
      assert_equal :mr_after_commit_call_procs_for_update, callback.args.first
    end

    should "add an after commit callback on destroy" do
      callback = subject.callbacks.find do |c|
        c.type == :after_commit && c.options[:on] == :destroy
      end

      assert_not_nil callback
      assert_equal :mr_after_commit_call_procs_for_destroy, callback.args.first
    end

  end

  class InitTests < MixinTests
    desc "and init"
    setup do
      @record = @record_class.new

      i = 0
      @record.add_after_commit_proc(:save){   @save_call_num   = i += 1 }
      @record.add_after_commit_proc(:create){ @create_call_num = i += 1 }
      @record.add_after_commit_proc(:update){ @update_call_num = i += 1 }

      @destroy_called = false
      @record.add_after_commit_proc(:destroy){ @destroy_called = true }
    end
    subject{ @record }

    should "call and clear its save and create procs after commit on create" do
      subject.instance_eval{ mr_after_commit_call_procs_for_create }

      assert_equal 1, @create_call_num
      assert_equal 2, @save_call_num
      assert_nil @update_call_num
      assert_false @destroy_called

      assert_empty subject.after_commit_procs(:create, :save)
      assert_not_empty subject.after_commit_procs(:update, :destroy)

      assert_equal 1, subject.called_after_commit_procs(:create).size
      assert_equal 1, subject.called_after_commit_procs(:save).size
      assert_equal 0, subject.called_after_commit_procs(:update).size
      assert_equal 0, subject.called_after_commit_procs(:destroy).size
    end

    should "call and clear its save and update procs after commit on update" do
      subject.instance_eval{ mr_after_commit_call_procs_for_update }

      assert_equal 1, @update_call_num
      assert_equal 2, @save_call_num
      assert_nil @create_call_num
      assert_false @destroy_called

      assert_empty subject.after_commit_procs(:update, :save)
      assert_not_empty subject.after_commit_procs(:create, :destroy)

      assert_equal 1, subject.called_after_commit_procs(:update).size
      assert_equal 1, subject.called_after_commit_procs(:save).size
      assert_equal 0, subject.called_after_commit_procs(:create).size
      assert_equal 0, subject.called_after_commit_procs(:destroy).size
    end

    should "call and clear its destroy procs after commit on destroy" do
      subject.instance_eval{ mr_after_commit_call_procs_for_destroy }

      assert_true @destroy_called
      assert_nil @save_call_num
      assert_nil @create_call_num
      assert_nil @update_call_num

      assert_empty subject.after_commit_procs(:destroy)
      assert_not_empty subject.after_commit_procs(:create, :update, :save)

      assert_equal 1, subject.called_after_commit_procs(:destroy).size
      assert_equal 0, subject.called_after_commit_procs(:update).size
      assert_equal 0, subject.called_after_commit_procs(:save).size
      assert_equal 0, subject.called_after_commit_procs(:create).size
    end

  end

end
