require 'assert'
require 'mr/factory/model_factory'

require 'mr/factory/config'
require 'mr/factory/record_factory'
require 'mr/fake_record'
require 'mr/test_helpers'
require 'mr/model'
require 'mr/record'

class MR::Factory::ModelFactory

  class UnitTests < Assert::Context
    include MR::TestHelpers

    desc "MR::Factory::ModelFactory"
    setup do
      @factory_class = MR::Factory::ModelFactory
    end
    subject{ @factory_class }

  end

  class InitTests < UnitTests
    desc "when init"
    setup do
      @model_class  = FakeTwoModel
      @record_class = @model_class.record_class

      @record_factory = nil
      Assert.stub(MR::Factory::RecordFactory, :new) do |*args|
        @record_factory ||= Assert.stub_send(MR::Factory::RecordFactory, :new, *args)
      end

      @factory = @factory_class.new(@model_class, @record_class)
    end
    subject{ @factory }

    should have_readers :model_class, :record_class, :config
    should have_imeths :model, :instance
    should have_imeths :saved_model, :saved_instance
    should have_imeths :stack, :instance_stack
    should have_imeths :stack_model, :saved_stack, :saved_stack_model
    should have_imeths :saved_dependencies_stack, :saved_deps_stack
    should have_imeths :saved_dependencies_stack_model, :saved_deps_stack_model

    should "instance eval any block it's initialized with" do
      yielded = nil
      factory = @factory_class.new(@model_class, @record_class){ yielded = self }
      assert_equal factory, yielded
    end

    should "know its model class, record class, and config" do
      assert_equal @model_class,  subject.model_class
      assert_equal @record_class, subject.record_class

      assert_instance_of Config, subject.config
      assert_equal subject.record_class, subject.config.record_class
    end

    should "build a record factory for its record class" do
      assert_not_nil @record_factory
      assert_equal @record_class, @record_factory.record_class
    end

    should "allow configuring a default association" do
      associated_factory = @factory_class.new(FakeOneModel, FakeOneRecord)
      association_name   = Factory.boolean ? :fake_one : :fake_poly
      factory = @factory_class.new(@model_class, @record_class) do
        default_association(association_name, associated_factory, {
          :force_in_stack => true
        })
      end

      associated_model = associated_factory.model
      Assert.stub(associated_factory, :model){ associated_model }

      result = factory.config.build_associated_model(
        association_name,
        associated_factory.record_class
      )
      assert_same associated_model, result

      assert_includes association_name, factory.config.force_in_stack_association_names
    end

    should "allow configuring default args" do
      name              = Factory.string
      default_args_proc = proc{ set(:name, name) }
      factory           = @factory_class.new(@model_class, @record_class)

      set_default_args_called_with = nil
      Assert.stub(factory.config, :set_default_args) do |&block|
        set_default_args_called_with = block
        Assert.stub_send(factory.config, :set_default_args, &block)
      end

      factory.instance_eval{ default_args(&default_args_proc) }
      assert_equal default_args_proc, set_default_args_called_with

      model = factory.model
      assert_equal name, model.name
    end

  end

  class ModelMethodTests < InitTests
    desc "model"
    setup do
      @built_record = nil
      Assert.stub(@record_factory, :record) do |*args|
        @built_record = Assert.stub_send(@record_factory, :record, *args)
      end
    end

    should "return an instance of the model class" do
      model = subject.model
      assert_instance_of @model_class, model
      assert_same @built_record, model.record
    end

    should "be aliased" do
      model = subject.instance
      assert_instance_of @model_class, model
      assert_same @built_record, model.record
    end

    should "apply passed args to the model" do
      name  = Factory.string
      model = subject.model(:name => name)
      assert_equal name, model.name
    end

    should "apply default args to the model" do
      name = Factory.string
      subject.instance_eval{ default_args{ set(:name, name) } }
      model = subject.model
      assert_equal name, model.name
    end

    should "apply passed args over default args to the model" do
      name = Factory.string
      subject.instance_eval{ default_args{ set(:name, name) } }
      new_name = Factory.string
      model    = subject.model(:name => new_name)
      assert_equal new_name, model.name
    end

  end

  class SavedModelMethodTests < InitTests
    desc "saved_model"
    setup do
      @model_called_with = nil
      @built_model       = nil
      Assert.stub(subject, :model) do |*args|
        @model_called_with = args
        @built_model       = Assert.stub_send(subject, :model, *args)
      end
    end

    should "build an instance of the model class and save it" do
      saved_model = subject.saved_model
      assert_equal [nil],          @model_called_with
      assert_same  @built_model,   saved_model
      assert_false saved_model.new?
    end

    should "be aliased" do
      saved_model = subject.saved_instance
      assert_equal [nil],          @model_called_with
      assert_same  @built_model,   saved_model
      assert_false saved_model.new?
    end

    should "allow passing args when building a saved model" do
      args        = { :name => Factory.string }
      saved_model = subject.saved_model(args)
      assert_equal [args], @model_called_with
    end

    should "reset the fake models save called flag" do
      assert_not_model_saved subject.saved_model
    end

  end

  class StackMethodTests < InitTests
    desc "stack"
    setup do
      @model_called_with = nil
      @built_model       = nil
      Assert.stub(subject, :model) do |*args|
        @model_called_with = args
        @built_model       = Assert.stub_send(subject, :model, *args)
      end
    end

    should "build a model stack" do
      stack = subject.stack
      assert_instance_of MR::Factory::ModelStack, stack
      assert_equal [nil],          @model_called_with
      assert_same  @built_model,   stack.model
      assert_same  subject.config, stack.factory_config
    end

    should "be aliased" do
      stack = subject.instance_stack
      assert_instance_of MR::Factory::ModelStack, stack
      assert_equal [nil],          @model_called_with
      assert_same  @built_model,   stack.model
      assert_same  subject.config, stack.factory_config
    end

    should "allow passing args when building a stack" do
      args  = { :name => Factory.string }
      stack = subject.stack(args)
      assert_equal [args], @model_called_with
    end

  end

  class StackModelMethodTests < InitTests
    desc "stack_model"
    setup do
      @stack_called_with = nil
      @built_stack       = nil
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
      end
    end

    should "build a stack model" do
      model = subject.stack_model
      assert_equal [nil],              @stack_called_with
      assert_same  @built_stack.model, model
    end

    should "allow passing args when building a stack model" do
      args  = { :name => Factory.string }
      model = subject.stack_model(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedStackMethodTests < InitTests
    desc "saved_stack"
    setup do
      @stack_called_with   = nil
      @built_stack         = nil
      @stack_create_called = false
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
        Assert.stub(@built_stack, :create) do
          @stack_create_called = true
          Assert.stub_send(@built_stack, :create)
        end
        @built_stack
      end
    end

    should "build a saved stack" do
      saved_stack = subject.saved_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_stack
      assert_true @stack_create_called
    end

    should "allow passing args when building a saved stack model" do
      args        = { :name => Factory.string }
      saved_stack = subject.saved_stack(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedStackModelMethodTests < InitTests
    desc "saved_stack_model"
    setup do
      @saved_stack_called_with = nil
      @built_saved_stack       = nil
      Assert.stub(subject, :saved_stack) do |*args|
        @saved_stack_called_with = args
        @built_saved_stack       = Assert.stub_send(subject, :saved_stack, *args)
      end
    end

    should "build a saved stack model" do
      saved_model = subject.saved_stack_model
      assert_equal [nil],                    @saved_stack_called_with
      assert_same  @built_saved_stack.model, saved_model
    end

    should "allow passing args when building a saved stack model" do
      args        = { :name => Factory.string }
      saved_model = subject.saved_stack_model(args)
      assert_equal [args], @saved_stack_called_with
    end

  end

  class SavedDepsStackMethodTests < InitTests
    desc "saved_dependencies_stack"
    setup do
      @stack_called_with        = nil
      @built_stack              = nil
      @stack_create_deps_called = false
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
        Assert.stub(@built_stack, :create_deps) do
          @stack_create_deps_called = true
          Assert.stub_send(@built_stack, :create_deps)
        end
        @built_stack
      end
    end

    should "build a saved dependencies stack" do
      saved_deps_stack = subject.saved_dependencies_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_deps_stack
      assert_true @stack_create_deps_called
    end

    should "be aliased" do
      saved_deps_stack = subject.saved_deps_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_deps_stack
      assert_true @stack_create_deps_called
    end

    should "allow passing args when building a saved dependencies stack" do
      args             = { :name => Factory.string }
      saved_deps_stack = subject.saved_dependencies_stack(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedDepsStackModelTests < InitTests
    desc "saved_dependencies_stack_model"
    setup do
      @saved_deps_stack_called_with = nil
      @built_saved_deps_stack       = nil
      Assert.stub(subject, :saved_deps_stack) do |*args|
        @saved_deps_stack_called_with = args
        @built_saved_deps_stack = Assert.stub_send(subject, :saved_deps_stack, *args)
      end
    end

    should "know how to build a saved dependencies stack model" do
      saved_deps_model = subject.saved_dependencies_stack_model
      assert_equal [nil],                         @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.model, saved_deps_model
    end

    should "be aliased" do
      saved_deps_model = subject.saved_deps_stack_model
      assert_equal [nil],                         @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.model, saved_deps_model
    end

    should "allow passing args when building a saved dependencies stack model" do
      args             = { :name => Factory.string }
      saved_deps_model = subject.saved_dependencies_stack_model(args)
      assert_equal [args],                        @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.model, saved_deps_model
    end

  end

  class ConfigTests < UnitTests
    desc "Config"
    setup do
      @config_class = Config
    end
    subject{ @config_class }

    should "be a with-associations factory config" do
      assert_includes MR::Factory::WithAssociationsConfig, subject
    end

  end

  class ConfigInitTests < ConfigTests
    desc "when init"
    setup do
      @record_class = FakeTwoRecord
      @config       = @config_class.new(@record_class)
    end
    subject{ @config }

    should have_imeths :build_associated_model, :build_associated_record
    should have_imeths :ar_association_for

    should "know how to build an associated model and record" do
      factory_for_called_with = []
      built_factory           = nil
      model                   = nil
      Assert.stub(subject, :factory_for) do |*args|
        factory_for_called_with = args
        built_factory = Assert.stub_send(subject, :factory_for, *args)
        model = built_factory.model
        Assert.stub(built_factory, :model){ model }
        built_factory
      end

      association_name   = Factory.boolean ? :fake_one : :fake_poly
      assoc_record_class = FakeOneRecord
      result = subject.build_associated_record(association_name, assoc_record_class)

      assert_equal [association_name, assoc_record_class], factory_for_called_with
      assert_instance_of MR::Factory::ModelFactory, built_factory
      assert_equal assoc_record_class,             built_factory.record_class
      assert_equal assoc_record_class.model_class, built_factory.model_class

      assert_same model, subject.build_associated_model(
        association_name,
        assoc_record_class
      )
      assert_same model.record, subject.build_associated_record(
        association_name,
        assoc_record_class
      )
    end

    should "know how to get an activerecord association for a model" do
      model = @record_class.model_class.new

      exp = model.record.association(:fake_one)
      assert_equal exp, subject.ar_association_for(model, :fake_one)

      assert_nil subject.ar_association_for(model, Factory.string)
    end

  end

  class FakeOneRecord
    include MR::FakeRecord
  end

  class FakeTwoRecord
    include MR::FakeRecord

    attribute :name,           :string
    attribute :fake_poly_type, :string
    attribute :fake_poly_id,   :integer
    attribute :fake_one_id,    :integer

    belongs_to :fake_poly, :polymorphic => true
    belongs_to :fake_one,  :class_name  => FakeOneRecord.to_s
  end

  class FakeOneModel
    include MR::Model
    record_class FakeOneRecord
  end

  class FakeTwoModel
    include MR::Model
    record_class FakeTwoRecord

    field_accessor :name, :fake_poly_type, :fake_poly_id, :fake_one_id

    polymorphic_belongs_to :fake_poly
    belongs_to :fake_one
  end

end
