require 'assert'
require 'mr/factory/read_model_factory'

require 'mr/json_field'
require 'mr/read_model'

class MR::Factory::ReadModelFactory

  class UnitTests < Assert::Context
    desc "MR::Factory::ReadModelFactory"
    setup do
      @factory_class = MR::Factory::ReadModelFactory
    end
    subject{ @factory_class }

  end

  class InitTests < UnitTests
    desc "when init"
    setup do
      @read_model_class = TestReadModel
      @factory          = @factory_class.new(@read_model_class)
    end
    subject{ @factory }

    should have_readers :read_model_class, :config
    should have_imeths :read_model_data, :read_model, :instance

    should "instance eval any block it's initialized with" do
      yielded = nil
      factory = @factory_class.new(@read_model_class){ yielded = self }
      assert_equal factory, yielded
    end

    should "raise an argument error if init without a read model or read model struct" do
      assert_nothing_raised{ @factory_class.new(@read_model_class) }
      assert_nothing_raised{ @factory_class.new(@read_model_class::StructReadModel) }

      assert_raises(ArgumentError){ @factory_class.new(Factory.string) }
    end

    should "know its read model class and config" do
      assert_equal @read_model_class, subject.read_model_class

      assert_instance_of Config, subject.config
      assert_equal subject.read_model_class, subject.config.read_model_class
    end

    should "allow configuring default args" do
      name              = Factory.string
      default_args_proc = proc{ set(:name, name) }
      factory           = @factory_class.new(@read_model_class)

      set_default_args_called_with = nil
      Assert.stub(factory.config, :set_default_args) do |&block|
        set_default_args_called_with = block
        Assert.stub_send(factory.config, :set_default_args, &block)
      end

      factory.instance_eval{ default_args(&default_args_proc) }
      assert_equal default_args_proc, set_default_args_called_with

      read_model = factory.read_model
      assert_equal name, read_model.name
    end

  end

  class ReadModelDataTests < InitTests
    desc "read_model_data"

    should "build a read model data object" do
      read_model_data = subject.read_model_data
      assert_instance_of Data, read_model_data
      assert_kind_of String, read_model_data['name']
    end

    should "apply passed args to the read model data" do
      name            = Factory.string
      read_model_data = subject.read_model_data(:name => name)
      assert_equal name, read_model_data['name']
    end

    should "apply default args to the record" do
      name = Factory.string
      subject.config.set_default_args{ set(:name, name) }
      read_model_data = subject.read_model_data
      assert_equal name, read_model_data['name']
    end

    should "apply passed args over default args to the read model data" do
      name = Factory.string
      subject.config.set_default_args{ set(:name, name) }
      new_name        = Factory.string
      read_model_data = subject.read_model_data(:name => new_name)
      assert_equal new_name, read_model_data['name']
    end

  end

  class ReadModelMethodTests < InitTests
    desc "read_model"
    setup do
      @read_model_data_called_with = nil
      @built_read_model_data       = nil
      Assert.stub(subject, :read_model_data) do |*args|
        @read_model_data_called_with = args
        @built_read_model_data = Assert.stub_send(subject, :read_model_data, *args)
      end
    end

    should "build an instance of the read model clast" do
      read_model = subject.read_model
      assert_instance_of @read_model_class, read_model
      assert_same @built_read_model_data, read_model.instance_eval{ read_model_data }
    end

    should "be aliased" do
      read_model = subject.instance
      assert_instance_of @read_model_class, read_model
      assert_same @built_read_model_data, read_model.instance_eval{ read_model_data }
    end

    should "allow passing args when building a read model" do
      args         = { :name => Factory.string }
      saved_record = subject.read_model(args)
      assert_equal [args], @read_model_data_called_with
    end

  end

  class ConfigTests < UnitTests
    desc "Config"
    setup do
      @config_class = Config
    end
    subject{ @config_class }

    should "be a factory config" do
      assert_includes MR::Factory::Config, subject
    end

  end

  class ConfigInitTests < ConfigTests
    desc "when init"
    setup do
      @read_model_class = TestReadModel
      @config           = @config_class.new(@read_model_class)
    end
    subject{ @config }

    should have_imeths :read_model_class

    should "know its read model class" do
      assert_equal @read_model_class, subject.read_model_class
    end

    should "know how to apply field default args to a read model data" do
      read_model_data = @factory_class::Data.new
      subject.apply_args(read_model_data, {})

      assert_kind_of String, read_model_data['name']

      assert_instance_of Array, read_model_data['struct_list']
      assert_includes read_model_data['struct_list'].size, [1, 2, 3]

      read_model_data['struct_list'].each do |struct_list_data|
        assert_instance_of Data, struct_list_data
        assert_kind_of String, struct_list_data['name']
      end

      assert_instance_of Data, read_model_data['struct_obj']
      assert_kind_of String, read_model_data['struct_obj']['name']
    end

  end

  class TestReadModel
    include MR::ReadModel

    field :name, :string

    json_struct_obj :struct_obj, 'StructReadModel'

    json_struct_list :struct_list, 'StructReadModel'

    class StructReadModel
      include MR::ReadModelStruct

      field :name, :string
    end
  end

end
