require 'assert'
require 'mr/factory/record_factory'

require 'mr/factory/config'
require 'mr/fake_record'
require 'mr/record'

class MR::Factory::RecordFactory

  class UnitTests < Assert::Context
    desc "MR::Factory::RecordFactory"
    setup do
      @factory_class = MR::Factory::RecordFactory
    end
    subject{ @factory_class }

  end

  class InitTests < UnitTests
    desc "when init"
    setup do
      @record_class = FakeTwoRecord
      @factory      = @factory_class.new(@record_class)
    end
    subject{ @factory }

    should have_readers :record_class, :config
    should have_imeths :record, :instance
    should have_imeths :saved_record, :saved_instance
    should have_imeths :stack, :instance_stack
    should have_imeths :stack_record, :saved_stack, :saved_stack_record
    should have_imeths :saved_dependencies_stack, :saved_deps_stack
    should have_imeths :saved_dependencies_stack_record, :saved_deps_stack_record

    should "instance eval any block it's initialized with" do
      yielded = nil
      factory = @factory_class.new(@record_class){ yielded = self }
      assert_equal factory, yielded
    end

    should "know its record class and config" do
      assert_equal @record_class, subject.record_class

      assert_instance_of Config, subject.config
      assert_equal subject.record_class, subject.config.record_class
    end

    should "allow configuring a default association" do
      association_name   = Factory.boolean ? :fake_one : :fake_poly
      assoc_record_class = FakeOneRecord
      associated_factory = @factory_class.new(assoc_record_class)
      factory = @factory_class.new(@record_class) do
        default_association(association_name, associated_factory, {
          :force_in_stack => true
        })
      end

      associated_record = associated_factory.record
      Assert.stub(associated_factory, :record){ associated_record }

      assert_same associated_record, factory.config.build_associated_record(
        association_name,
        assoc_record_class
      )

      assert_includes association_name, factory.config.force_in_stack_association_names
    end

    should "allow configuring default args" do
      name              = Factory.string
      default_args_proc = proc{ set(:name, name) }
      factory           = @factory_class.new(@record_class)

      set_default_args_called_with = nil
      Assert.stub(factory.config, :set_default_args) do |&block|
        set_default_args_called_with = block
        Assert.stub_send(factory.config, :set_default_args, &block)
      end

      factory.instance_eval{ default_args(&default_args_proc) }
      assert_equal default_args_proc, set_default_args_called_with

      record = factory.record
      assert_equal name, record.name
    end

  end

  class RecordMethodTests < InitTests
    desc "record"
    setup do
      @built_record = nil
      Assert.stub(@record_class, :new) do |*args|
        @built_record = Assert.stub_send(@record_class, :new, *args)
      end
    end

    should "return an instance of the record class" do
      record = subject.record
      assert_same @built_record, record
      assert_instance_of String, record.name
    end

    should "be aliased" do
      record = subject.instance
      assert_same @built_record, record
      assert_instance_of String, record.name
    end

    should "apply passed args to the record" do
      name   = Factory.string
      record = subject.record(:name => name)
      assert_equal name, record.name
    end

    should "apply default args to the record" do
      name = Factory.string
      subject.config.set_default_args{ set(:name, name) }
      record = subject.record
      assert_equal name, record.name
    end

    should "apply passed args over default args to the record" do
      name = Factory.string
      subject.config.set_default_args{ set(:name, name) }
      new_name  = Factory.string
      record    = subject.record(:name => new_name)
      assert_equal new_name, record.name
    end

  end

  class SavedRecordMethodTests < InitTests
    desc "saved_record"
    setup do
      @record_called_with = nil
      @built_record       = nil
      Assert.stub(subject, :record) do |*args|
        @record_called_with = args
        @built_record       = Assert.stub_send(subject, :record, *args)
      end
    end

    should "build an instance of the record class and save it" do
      saved_record = subject.saved_record
      assert_same @built_record, saved_record
      assert_false saved_record.new_record?
    end

    should "be aliased" do
      saved_record = subject.saved_instance
      assert_same @built_record, saved_record
      assert_false saved_record.new_record?
    end

    should "allow passing args when building a saved record" do
      args         = { :name => Factory.string }
      saved_record = subject.saved_record(args)
      assert_equal [args], @record_called_with
    end

    should "reset the fake records save called flag" do
      assert_false subject.saved_record.save_called
    end

  end

  class StackMethodTests < InitTests
    desc "stack"
    setup do
      @record_called_with = nil
      @built_record       = nil
      Assert.stub(subject, :record) do |*args|
        @record_called_with = args
        @built_record       = Assert.stub_send(subject, :record, *args)
      end
    end

    should "build a record stack" do
      stack = subject.stack
      assert_instance_of MR::Factory::RecordStack, stack
      assert_equal [nil],          @record_called_with
      assert_same  @built_record,  stack.record
      assert_same  subject.config, stack.factory_config
    end

    should "be aliased" do
      stack = subject.instance_stack
      assert_instance_of MR::Factory::RecordStack, stack
      assert_equal [nil],          @record_called_with
      assert_same  @built_record,  stack.record
      assert_same  subject.config, stack.factory_config
    end

    should "allow passing args when building a stack" do
      args  = { :name => Factory.string }
      stack = subject.stack(args)
      assert_equal [args], @record_called_with
    end

  end

  class StackRecordMethodTests < InitTests
    desc "stack_record"
    setup do
      @stack_called_with = nil
      @built_stack       = nil
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
      end
    end

    should "build a stack record" do
      record = subject.stack_record
      assert_equal [nil],               @stack_called_with
      assert_same  @built_stack.record, record
    end

    should "allow passing args when building a stack record" do
      args   = { :name => Factory.string }
      record = subject.stack_record(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedStackMethodTests < InitTests
    desc "saved_stack"
    setup do
      @stack_called_with   = nil
      @built_stack         = nil
      @stack_create_called = false
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
        Assert.stub(@built_stack, :create) do
          @stack_create_called = true
          Assert.stub_send(@built_stack, :create)
        end
        @built_stack
      end
    end

    should "build a saved stack" do
      saved_stack = subject.saved_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_stack
      assert_true @stack_create_called
    end

    should "allow passing args when building a saved stack record" do
      args        = { :name => Factory.string }
      saved_stack = subject.saved_stack(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedStackModelMethodTests < InitTests
    desc "saved_stack_record"
    setup do
      @saved_stack_called_with = nil
      @built_saved_stack       = nil
      Assert.stub(subject, :saved_stack) do |*args|
        @saved_stack_called_with = args
        @built_saved_stack       = Assert.stub_send(subject, :saved_stack, *args)
      end
    end

    should "build a saved stack record" do
      saved_record = subject.saved_stack_record
      assert_equal [nil],                     @saved_stack_called_with
      assert_same  @built_saved_stack.record, saved_record
    end

    should "allow passing args when building a saved stack record" do
      args         = { :name => Factory.string }
      saved_record = subject.saved_stack_record(args)
      assert_equal [args], @saved_stack_called_with
    end

  end

  class SavedDepsStackMethodTests < InitTests
    desc "saved_dependencies_stack"
    setup do
      @stack_called_with        = nil
      @built_stack              = nil
      @stack_create_deps_called = false
      Assert.stub(subject, :stack) do |*args|
        @stack_called_with = args
        @built_stack       = Assert.stub_send(subject, :stack, *args)
        Assert.stub(@built_stack, :create_deps) do
          @stack_create_deps_called = true
          Assert.stub_send(@built_stack, :create_deps)
        end
        @built_stack
      end
    end

    should "build a saved dependencies stack" do
      saved_deps_stack = subject.saved_dependencies_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_deps_stack
      assert_true @stack_create_deps_called
    end

    should "be aliased" do
      saved_deps_stack = subject.saved_deps_stack
      assert_equal [nil],        @stack_called_with
      assert_same  @built_stack, saved_deps_stack
      assert_true @stack_create_deps_called
    end

    should "allow passing args when building a saved dependencies stack" do
      args             = { :name => Factory.string }
      saved_deps_stack = subject.saved_dependencies_stack(args)
      assert_equal [args], @stack_called_with
    end

  end

  class SavedDepsStackModelTests < InitTests
    desc "saved_dependencies_stack_record"
    setup do
      @saved_deps_stack_called_with = nil
      @built_saved_deps_stack       = nil
      Assert.stub(subject, :saved_deps_stack) do |*args|
        @saved_deps_stack_called_with = args
        @built_saved_deps_stack = Assert.stub_send(subject, :saved_deps_stack, *args)
      end
    end

    should "know how to build a saved dependencies stack record" do
      saved_deps_record = subject.saved_dependencies_stack_record
      assert_equal [nil],                          @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.record, saved_deps_record
    end

    should "be aliased" do
      saved_deps_record = subject.saved_deps_stack_record
      assert_equal [nil],                          @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.record, saved_deps_record
    end

    should "allow passing args when building a saved dependencies stack record" do
      args              = { :name => Factory.string }
      saved_deps_record = subject.saved_dependencies_stack_record(args)
      assert_equal [args],                         @saved_deps_stack_called_with
      assert_same  @built_saved_deps_stack.record, saved_deps_record
    end

  end

  class ConfigTests < UnitTests
    desc "Config"
    setup do
      @config_class = Config
    end
    subject{ @config_class }

    should "be a with-associations factory config" do
      assert_includes MR::Factory::WithAssociationsConfig, subject
    end

  end

  class ConfigInitTests < ConfigTests
    desc "when init"
    setup do
      @record_class = FakeTwoRecord
      @config       = @config_class.new(@record_class)
    end
    subject{ @config }

    should have_imeths :build_associated_record, :ar_association_for

    should "know how to build an associated record" do
      factory_for_called_with = []
      built_factory           = nil
      record                  = nil
      Assert.stub(subject, :factory_for) do |*args|
        factory_for_called_with = args
        built_factory = Assert.stub_send(subject, :factory_for, *args)
        record = built_factory.record
        Assert.stub(built_factory, :record){ record }
        built_factory
      end

      association_name   = Factory.boolean ? :fake_one : :fake_poly
      assoc_record_class = FakeOneRecord
      result = subject.build_associated_record(
        association_name,
        assoc_record_class
      )

      assert_equal [association_name, assoc_record_class], factory_for_called_with
      assert_instance_of MR::Factory::RecordFactory, built_factory
      assert_equal assoc_record_class, built_factory.record_class

      assert_same record, result
    end

    should "know how to get an activerecord association for a record" do
      record = @record_class.new

      exp = record.association(:fake_one)
      assert_equal exp, subject.ar_association_for(record, :fake_one)

      assert_nil subject.ar_association_for(record, Factory.string)
    end

    should "know how to apply column default args to a record" do
      record = @record_class.new
      subject.apply_args(record, {})

      assert_instance_of String, record.name
      assert_nil record.id
      assert_nil record.fake_poly_type
      assert_nil record.fake_poly_id
      assert_nil record.fake_one_id
      assert_nil record.bad_column
    end

  end

  class FakeOneRecord
    include MR::FakeRecord
  end

  class FakeTwoRecord
    include MR::FakeRecord

    attribute :name,           :string
    attribute :fake_poly_type, :string
    attribute :fake_poly_id,   :integer
    attribute :fake_one_id,    :integer

    attribute :bad_column, :invalid_type

    belongs_to :fake_poly, :polymorphic => true
    belongs_to :fake_one,  :class_name  => FakeOneRecord.to_s
  end

end
