require 'assert'
require 'mr/json_field'

require 'much-plugin'

module MR::JsonField

  class UnitTests < Assert::Context
    desc "MR::JsonField"
    setup do
      @module = MR::JsonField

      @orig_encoder = @module.encoder
      @orig_decoder = @module.decoder
    end
    teardown do
      @module.encoder = @orig_encoder
      @module.decoder = @orig_decoder
    end
    subject{ @module }

    should have_imeths :encoder, :encoder=
    should have_imeths :decoder, :decoder=
    should have_imeths :encode, :decode

    should "use much-plugin" do
      assert_includes MuchPlugin, subject
    end

    should "know its default encoder and decoder" do
      value = { Factory.string => Factory.string }

      exp = JSON.dump(value)
      assert_equal exp, DEFAULT_ENCODER.call(value)

      encoded_value = exp
      exp = JSON.load(encoded_value)
      assert_equal exp, DEFAULT_DECODER.call(encoded_value)
    end

    should "default its encoder and decoder" do
      assert_equal DEFAULT_ENCODER, subject.encoder
      assert_equal DEFAULT_DECODER, subject.decoder
    end

    should "use its encoder/decoder when encoding/decoding" do
      value = { Factory.string => Factory.string }

      exp = subject.encoder.call(value)
      assert_equal exp, subject.encode(value)

      encoded_value = exp
      exp = subject.decoder.call(encoded_value)
      assert_equal exp, subject.decode(encoded_value)
    end

    should "allow specifying a custom encoder/decoder" do
      subject.encoder = proc{ |v| v.to_s }
      subject.decoder = proc{ |v| v.to_i }

      value = Factory.integer
      assert_equal value.to_s, subject.encode(value)

      value = Factory.integer.to_s
      assert_equal value.to_i, subject.decode(value)
    end

  end

  class MixedInSetupTests < UnitTests
    setup do
      @json_field_class = Class.new do
        include MR::JsonField
        attr_accessor :test_data_json, :custom_data_source
      end
    end

  end

  class MixedInTests < MixedInSetupTests
    desc "when mixed in"
    subject{ @json_field_class }

    should have_imeths :json_field, :json_field_reader, :json_field_writer
    should have_imeths :json_field_readers, :json_field_writers
    should have_imeths :json_field_accessors, :json_field_source_fields

    should "not have any json field accessors or source fields by default" do
      assert_equal [], subject.json_field_readers
      assert_equal [], subject.json_field_writers
      assert_equal [], subject.json_field_accessors
      assert_equal [], subject.json_field_source_fields
    end

    should "allow adding a json field reader" do
      subject.json_field_reader 'test_data'
      instance = subject.new

      assert_respond_to 'test_data', instance
      assert_equal ['test_data'],      subject.json_field_readers
      assert_equal ['test_data_json'], subject.json_field_source_fields
      assert_equal [], subject.json_field_writers
      assert_equal [], subject.json_field_accessors
    end

    should "allow adding a json field reader with a custom source" do
      subject.json_field_reader 'custom_data', :source => 'custom_data_source'
      instance = subject.new

      assert_respond_to 'custom_data', instance
      assert_equal ['custom_data'],        subject.json_field_readers
      assert_equal ['custom_data_source'], subject.json_field_source_fields
      assert_equal [], subject.json_field_writers
      assert_equal [], subject.json_field_accessors
    end

    should "allow adding a json field writer" do
      subject.json_field_writer 'test_data'
      instance = subject.new

      assert_respond_to 'test_data=', instance
      assert_equal ['test_data'],      subject.json_field_writers
      assert_equal ['test_data_json'], subject.json_field_source_fields
      assert_equal [], subject.json_field_readers
      assert_equal [], subject.json_field_accessors
    end

    should "allow adding a json field writer with a custom source" do
      subject.json_field_writer 'custom_data', :source => 'custom_data_source'
      instance = subject.new

      assert_respond_to 'custom_data=', instance
      assert_equal ['custom_data'],        subject.json_field_writers
      assert_equal ['custom_data_source'], subject.json_field_source_fields
      assert_equal [], subject.json_field_readers
      assert_equal [], subject.json_field_accessors
    end

    should "allow adding a json field accessor" do
      subject.json_field 'test_data'
      instance = subject.new

      assert_respond_to 'test_data',  instance
      assert_respond_to 'test_data=', instance
      exp = ['test_data']
      assert_equal exp, subject.json_field_readers
      assert_equal exp, subject.json_field_writers
      assert_equal exp, subject.json_field_accessors
      assert_equal ['test_data_json'], subject.json_field_source_fields
    end

    should "allow adding a json field accessor with a custom source" do
      subject.json_field 'custom_data', :source => 'custom_data_source'
      instance = subject.new

      assert_respond_to 'custom_data',  instance
      assert_respond_to 'custom_data=', instance
      exp = ['custom_data']
      assert_equal exp, subject.json_field_readers
      assert_equal exp, subject.json_field_writers
      assert_equal exp, subject.json_field_accessors
      assert_equal ['custom_data_source'], subject.json_field_source_fields
    end

    should "know it has an accessor if a reader and writer are added separately" do
      subject.json_field_reader 'test_data'
      subject.json_field_writer 'test_data'
      instance = subject.new

      assert_equal ['test_data'],      subject.json_field_accessors
      assert_equal ['test_data_json'], subject.json_field_source_fields
    end

    should "allow using symbols when adding json fields" do
      subject.json_field :test_data
      subject.json_field_reader :custom_data, :source => :custom_data_source
      subject.json_field_writer :custom_data, :source => :custom_data_source
      instance = subject.new

      assert_respond_to :test_data,    instance
      assert_respond_to :test_data=,   instance
      assert_respond_to :custom_data,  instance
      assert_respond_to :custom_data=, instance
      exp = ['test_data', 'custom_data']
      assert_equal exp, subject.json_field_readers
      assert_equal exp, subject.json_field_writers
      assert_equal exp, subject.json_field_accessors
      exp = ['test_data_json', 'custom_data_source']
      assert_equal exp, subject.json_field_source_fields
    end

    should "error if the source matches the field name" do
      name = Factory.string
      assert_raises(ArgumentError) do
        subject.json_field_reader(name, :source => name)
      end
      assert_raises(ArgumentError) do
        subject.json_field_writer(name, :source => name)
      end
      assert_raises(ArgumentError) do
        subject.json_field(name, :source => name)
      end
    end

  end

  class InitTests < MixedInTests
    desc "and init"
    setup do
      @value = [
        Factory.string,
        Factory.integer,
        Factory.boolean,
        Factory.integer(3).times{ Factory.integer },
        { Factory.string => Factory.string }
      ].sample

      @json_field_class.json_field :test_data
      @json_field_class.json_field :custom_data, :source => :custom_data_source

      @instance = @json_field_class.new
    end
    subject{ @instance }

    should "allow reading/writing json fields" do
      assert_nil subject.test_data
      assert_nil subject.test_data_json
      assert_nil subject.custom_data
      assert_nil subject.custom_data_source

      subject.test_data   = @value
      subject.custom_data = @value

      exp = MR::JsonField.encode(@value)
      assert_equal exp, subject.test_data_json
      assert_equal exp, subject.custom_data_source
      assert_equal @value, subject.test_data
      assert_equal @value, subject.custom_data
    end

    should "cache read/written values" do
      subject.test_data   = @value
      subject.custom_data = @value

      result = subject.test_data
      assert_same result, subject.test_data
      result = subject.custom_data
      assert_same result, subject.custom_data
    end

    should "allow reading/writing `nil` values" do
      subject.test_data   = @value
      subject.custom_data = @value

      assert_nothing_raised{ subject.test_data = nil }
      assert_nil subject.test_data
      assert_nothing_raised{ subject.custom_data = nil }
      assert_nil subject.custom_data
    end

    should "raise an invalid json error when reading invalid JSON" do
      subject.test_data_json     = Factory.string
      subject.custom_data_source = Factory.string

      assert_raises(InvalidJSONError){ subject.test_data }
      assert_raises(InvalidJSONError){ subject.custom_data }
    end

    should "raise an invalid json error when writing invalid JSON" do
      Assert.stub(MR::JsonField.encoder, :call){ raise RuntimeError }

      assert_raises(InvalidJSONError){ subject.test_data   = Factory.string }
      assert_raises(InvalidJSONError){ subject.custom_data = Factory.string }
    end

  end

  class TestHelpersTests < MixedInSetupTests
    include MR::JsonField::TestHelpers

    desc "TestHelpers"
    setup do
      @json_field_class.json_field :test_data
      @json_field_class.json_field :custom_data, :source => :custom_data_source

      @instance = @json_field_class.new
    end
    subject{ @instance }

    should "provide helpers for testing that a class has json fields" do
      assert_json_field subject, :test_data
      assert_json_field_reader subject, :test_data
      assert_json_field_writer subject, :test_data

      assert_json_field subject, :custom_data, :source => :custom_data_source
      assert_json_field_reader subject, :custom_data, :source => :custom_data_source
      assert_json_field_writer subject, :custom_data, :source => :custom_data_source
    end

  end

end
