require 'assert'
require 'mr/read_model/query_expression'

require 'ardb/relation_spy'
require 'mr/fake_record'
require 'mr/read_model/set_querying'
require 'test/support/read_model/querying'

module MR::ReadModel

  class UnitTests < Assert::Context
    include Querying::TestHelpers
    desc "MR::ReadModel"

  end

  class QueryExpressionTests < UnitTests
    desc "QueryExpression"
    subject{ QueryExpression }

    should "return the correct type of expression using `new`" do
      expression = subject.new(:select, 'column')
      assert_instance_of StaticQueryExpression, expression
      expression = subject.new(:select){ 'column' }
      assert_instance_of DynamicQueryExpression, expression
    end

    should "raise an invalid error when not passed args or a block" do
      assert_raises(InvalidQueryExpressionError){ subject.new(:select) }
    end

  end

  class StaticQueryExpressionTests < UnitTests
    desc "StaticQueryExpression"
    setup do
      @ar_relation = FakeTestRecord.scoped
      @expression  = StaticQueryExpression.new(:select, 'column')
    end
    subject{ @expression }

    should have_readers :type, :args

    should "apply itself to an ActiveRecord relation using `apply_to`" do
      subject.apply_to(@ar_relation)
      assert_expression_applied @ar_relation, subject.type, *subject.args
    end

  end

  class DynamicQueryExpressionTests < UnitTests
    desc "DynamicQueryExpression"
    setup do
      @ar_relation = FakeTestRecord.scoped
      block = proc{ 'column' }
      @expression  = DynamicQueryExpression.new(:select, &block)
    end
    subject{ @expression }

    should have_readers :type, :block

    should "apply itself to an ActiveRecord relation using `apply_to`" do
      subject.apply_to(@ar_relation, 'test')
      assert_expression_applied @ar_relation, subject.type, 'column'
    end

    should "yield any params to it's block using `apply_to`" do
      yielded = nil
      block = proc{ |params| yielded = params }
      expression = DynamicQueryExpression.new(:select, &block)
      expression.apply_to(@ar_relation, 'test')
      assert_equal 'test', yielded
    end

    should "eval it's block in the ActiveRecord relation's scope using `apply_to`" do
      scope = nil
      block = proc{ |params| scope = self }
      expression = DynamicQueryExpression.new(:select, &block)
      expression.apply_to(@ar_relation, 'test')
      assert_equal @ar_relation, scope
    end

  end

  class MergeQueryExpressionTests < UnitTests
    desc "MergeQueryExpression"
    setup do
      @ar_relation = FakeTestRecord.scoped
      @expression  = MergeQueryExpression.new(:order, 'relation')
    end
    subject{ @expression }

    should have_readers :type, :query_expression
    should have_imeths :apply_to

    should "build a query expression for a merge" do
      query_expression = subject.query_expression
      assert_instance_of StaticQueryExpression, query_expression
      assert_equal :merge, query_expression.type
      assert_equal [ 'relation' ], query_expression.args
    end

    should "apply it's query expression using `apply_to`" do
      query_expression = subject.query_expression
      subject.apply_to(@ar_relation)
      assert_expression_applied @ar_relation, query_expression.type, 'relation'
    end

  end

  class SubqueryExpressionTests < UnitTests
    desc "SubqueryExpression"
    setup do
      @expression_class = SubqueryExpression
    end
    subject{ @expression_class }

    should "know its types" do
      exp = {
        :joins => JoinSubquery
      }
      assert_equal exp, subject::TYPES
    end

  end

  class SubqueryExpressionInitTests < SubqueryExpressionTests
    desc "when init"
    setup do
      @type  = @expression_class::TYPES.keys.sample
      @args  = Factory.integer(3).times.map{ Factory.string }
      @block = proc{ Factory.string }

      subquery_class = @expression_class::TYPES[@type]
      Assert.stub(subquery_class, :new) do |*args, &block|
        SubquerySpy.new(*args, &block)
      end

      @expression = @expression_class.new(@type, *@args, &@block)
    end
    subject{ @expression }

    should have_readers :subquery_type, :subquery_args, :subquery_block
    should have_imeths :type, :subquery

    should "know its subquery attributes" do
      assert_equal @type,  subject.subquery_type
      assert_equal @args,  subject.subquery_args
      assert_equal @block, subject.subquery_block
    end

    should "alias its subquery type, to be a compatible query expression" do
      assert_equal subject.subquery_type, subject.type
    end

    should "know its subquery" do
      subquery_spy = subject.subquery
      assert_instance_of SubquerySpy, subquery_spy
      assert_equal subject.subquery_args,  subquery_spy.args
      assert_equal subject.subquery_block, subquery_spy.block
      assert_same subquery_spy, subject.subquery
    end

    should "apply the subquery's SQL to a relation using `apply_to`" do
      ar_relation = FakeTestRecord.scoped
      subject.apply_to(ar_relation)

      subquery_spy = subject.subquery
      assert_nil subquery_spy.sql_built_with
      assert_expression_applied ar_relation, subject.type, subquery_spy.built_sql

      params = { Factory.string => Factory.string }
      ar_relation = FakeTestRecord.scoped
      subject.apply_to(ar_relation, params)

      subquery_spy = subject.subquery
      assert_equal params, subquery_spy.sql_built_with
      assert_expression_applied ar_relation, subject.type, subquery_spy.built_sql
    end

    should "raise an argument error if a block isn't passed" do
      assert_raises(ArgumentError){ @expression_class.new(@type, *@args) }
    end

  end

  class FromExpressionTests < UnitTests
    desc "FromExpression"
    setup do
      @expression_class = FromExpression
      @from_expression  = @expression_class.new(FakeTestRecord)
    end
    subject{ @from_expression }

    should have_readers :record_class
    should have_imeths :ar_relation

    should "know its record class" do
      assert_equal FakeTestRecord, subject.record_class
    end

    should "use its record class to build an ar relation" do
      assert_instance_of Ardb::RelationSpy, subject.ar_relation
      params = { Factory.string => Factory.string }
      assert_instance_of Ardb::RelationSpy, subject.ar_relation(params)
    end

    should "raise an argument error if passed a non MR record class" do
      assert_raises(ArgumentError){ @expression_class.new(Class.new) }
    end

  end

  class FromSubqueryExpressionTests < UnitTests
    desc "FromSubqueryExpression"
    setup do
      @expression_class = FromSubqueryExpression
      @subquery_block   = proc do
        read_model{ from FakeTestRecord }
      end
      @from_expression = @expression_class.new(&@subquery_block)
    end
    subject{ @from_expression }

    should have_readers :subquery_block
    should have_imeths :from_subquery, :record_class, :ar_relation

    should "know its subquery block" do
      assert_equal @subquery_block, subject.subquery_block
    end

    should "know its from subquery" do
      subquery = subject.from_subquery
      assert_instance_of FromSubquery, subquery
      assert_equal FakeTestRecord, subquery.record_class
      assert_same subquery, subject.from_subquery
    end

    should "know its record class" do
      assert_equal subject.from_subquery.record_class, subject.record_class
    end

    should "use its from subquery to build an ar relation" do
      build_sql_called_with = nil
      from_subquery_sql     = Factory.text
      Assert.stub(subject.from_subquery, :build_sql) do |params|
        build_sql_called_with = params
        from_subquery_sql
      end

      ar_relation_spy = subject.ar_relation
      assert_instance_of Ardb::RelationSpy, ar_relation_spy
      assert_nil build_sql_called_with
      assert_expression_applied ar_relation_spy, :from, from_subquery_sql

      params = { Factory.string => Factory.string }
      ar_relation_spy = subject.ar_relation(params)
      assert_instance_of Ardb::RelationSpy, ar_relation_spy
      assert_equal params, build_sql_called_with
      assert_expression_applied ar_relation_spy, :from, from_subquery_sql
    end

    should "raise an argument error if not passed a block" do
      assert_raises(ArgumentError){ @expression_class.new }
    end

  end

  class NullFromExpressionTests < UnitTests
    desc "NullFromExpression"
    setup do
      @from_expression = NullFromExpression.new
    end
    subject{ @from_expression }

    should have_imeths :record_class, :ar_relation

    should "raise a no from expression error" do
      assert_raises(NoFromExpressionError){ subject.record_class }
      assert_raises(NoFromExpressionError){ subject.ar_relation }
      assert_raises(NoFromExpressionError) do
        subject.ar_relation(Factory.string => Factory.string)
      end
    end

  end

  class SetExpressionTests < UnitTests
    desc "SetExpression"
    setup do
      @expression_class = SetExpression
    end
    subject{ @expression_class }

    should "know its operator SQL" do
      exp = {
        :union         => 'UNION',
        :union_all     => 'UNION ALL',
        :intersect     => 'INTERSECT',
        :intersect_all => 'INTERSECT ALL',
        :except        => 'EXCEPT',
        :except_all    => 'EXCEPT ALL'
      }
      assert_equal exp, subject::OPERATOR_SQL
    end

  end

  class SetExpressionInitTests < SetExpressionTests
    desc "when init"
    setup do
      @operator = @expression_class::OPERATOR_SQL.keys.sample
      column = Factory.string
      @block = proc do
        select{ |p| p ? p['column'] : column }
        from FakeTestRecord
      end
      @expression = @expression_class.new(@operator, &@block)
    end
    subject{ @expression }

    should have_readers :operator_sql, :read_model_block
    should have_imeths :read_model_class, :combine_sql

    should "know its attributes" do
      exp = @expression_class::OPERATOR_SQL[@operator]
      assert_equal exp,    subject.operator_sql
      assert_equal @block, subject.read_model_block
    end

    should "know its read model class" do
      read_model_class = subject.read_model_class
      assert_includes MR::ReadModel,              read_model_class
      assert_includes MR::ReadModel::SetQuerying, read_model_class
      assert_equal FakeTestRecord, read_model_class.record_class
      assert_same read_model_class, subject.read_model_class
    end

    should "combine SQL with its operator using `combine_sql`" do
      sql = Factory.text
      exp = "#{sql} #{subject.operator_sql} " \
            "#{subject.read_model_class.relation.build_for_all.to_sql.strip}"
      assert_equal exp, subject.combine_sql(sql)

      params = { 'column' => Factory.string }
      exp = "#{sql} #{subject.operator_sql} " \
            "#{subject.read_model_class.build_sql(params)}"
      assert_equal exp, subject.combine_sql(sql, params)
    end

    should "raise an argument error if a block isn't passed" do
      assert_raises(ArgumentError){ @expression_class.new(@type, *@args) }
    end

  end

  class FakeTestRecord
    include MR::FakeRecord

    def self.scoped
      Ardb::RelationSpy.new
    end
  end

  class SubquerySpy
    attr_reader :args, :block, :built_sql, :sql_built_with

    def initialize(*args, &block)
      @args           = args
      @block          = block
      @built_sql      = Factory.text
      @sql_built_with = nil
    end

    def build_sql(params = nil)
      @sql_built_with = params
      @built_sql
    end
  end

end
