require 'assert'
require 'mr/read_model/set_querying'

require 'ardb/relation_spy'
require 'much-plugin'
require 'mr/fake_record'
require 'mr/read_model/query_expression'
require 'test/support/read_model/querying'

module MR::ReadModel::SetQuerying

  class UnitTests < Assert::Context
    include MR::ReadModel::Querying::TestHelpers

    desc "MR::ReadModel::SetQuerying"
    setup do
      @read_model_class = Class.new do
        include MR::ReadModel::SetQuerying
      end
    end
    subject{ @read_model_class }

    should have_imeths :relation
    should have_imeths :union, :union_all
    should have_imeths :intersect, :intersect_all
    should have_imeths :except, :except_all

    should "use much-plugin" do
      assert_includes MuchPlugin, MR::ReadModel::SetQuerying
    end

    should "mixin read model querying mixin" do
      assert_includes MR::ReadModel::Querying, subject
    end

    should "return a set querying relation using `relation`" do
      relation = subject.relation
      assert_instance_of MR::ReadModel::SetQuerying::Relation, relation
      assert_same relation, subject.relation
    end

  end

  class WithFromRecordClassTests < UnitTests
    setup do
      @ar_relation_spy = Ardb::RelationSpy.new
      Assert.stub(FakeTestRecord, :scoped){ @ar_relation_spy }
      @read_model_class.from FakeTestRecord
      @relation = @read_model_class.relation
    end

    should "add a union to the relation using `union`" do
      union_proc = proc{ read_model FakeTestRecord }
      subject.union(&union_proc)
      assert_set_expression_added @relation, :union, union_proc
    end

    should "add a union all to the relation using `union_all`" do
      union_all_proc = proc{ read_model FakeTestRecord }
      subject.union_all(&union_all_proc)
      assert_set_expression_added @relation, :union_all, union_all_proc
    end

    should "add a intersect to the relation using `intersect`" do
      intersect_proc = proc{ read_model FakeTestRecord }
      subject.intersect(&intersect_proc)
      assert_set_expression_added @relation, :intersect, intersect_proc
    end

    should "add a intersect all to the relation using `intersect_all`" do
      intersect_all_proc = proc{ read_model FakeTestRecord }
      subject.intersect_all(&intersect_all_proc)
      assert_set_expression_added @relation, :intersect_all, intersect_all_proc
    end

    should "add a except to the relation using `except`" do
      except_proc = proc{ read_model FakeTestRecord }
      subject.except(&except_proc)
      assert_set_expression_added @relation, :except, except_proc
    end

    should "add a except all to the relation using `except_all`" do
      except_all_proc = proc{ read_model FakeTestRecord }
      subject.except_all(&except_all_proc)
      assert_set_expression_added @relation, :except_all, except_all_proc
    end

    should "raise an argument error when any set method isn't provided a block" do
      assert_raises(ArgumentError){ subject.union }
      assert_raises(ArgumentError){ subject.union_all }
      assert_raises(ArgumentError){ subject.intersect }
      assert_raises(ArgumentError){ subject.intersect_all }
      assert_raises(ArgumentError){ subject.except }
      assert_raises(ArgumentError){ subject.except_all }
    end

  end

  class RelationTests < UnitTests
    desc "Relation"
    setup do
      @relation_class = Relation
      @relation       = @relation_class.new
    end
    subject{ @relation }

    should have_readers :set_expressions
    should have_imeths :build_sql

    should "not have any set expressions by default" do
      assert_equal [], subject.set_expressions
    end

    should "know how to build its sql" do
      Factory.integer(3).times.each do
        column = Factory.string

        @relation.set_expressions << Factory.read_model_set_expression do
          select{ |p| p ? p['column'] : column }
          from FakeTestRecord
        end
      end

      built_sql    = Factory.text
      relation_spy = Ardb::RelationSpy.new
      Assert.stub(relation_spy, :to_sql){ built_sql }
      Assert.stub(subject, :build_for_all){ relation_spy }

      exp = @relation.set_expressions.inject(built_sql){ |s, e| e.combine_sql(s) }
      assert_equal exp, subject.build_sql

      params = { Factory.string => Factory.string }
      exp = @relation.set_expressions.inject(built_sql) do |s, e|
        e.combine_sql(s, params)
      end
      assert_equal exp, subject.build_sql(params)
    end

  end

  class FakeTestRecord
    include MR::FakeRecord

    def self.scoped
      Ardb::RelationSpy.new
    end
  end

end
