require 'assert'
require 'mr/read_model'

require 'much-plugin'
require 'mr/read_model/data'
require 'mr/read_model/fields'
require 'mr/read_model/querying'

module MR::ReadModel

  class UnitTests < Assert::Context
    desc "MR::ReadModel"
    setup do
      @read_model_class = Class.new do
        include MR::ReadModel
      end
    end
    subject{ MR::ReadModel }

    should have_imeths :add_select_for_field

    should "use much-plugin" do
      assert_includes MuchPlugin, MR::ReadModel
    end

    should "know how to add a select for a field with column SQL" do
      field_name = Factory.string
      column_sql = Factory.string
      MR::ReadModel.add_select_for_field(@read_model_class, field_name, column_sql)

      assert_equal 1, @read_model_class.relation.query_expressions.size
      expression = @read_model_class.relation.query_expressions.first
      assert_equal :select, expression.type
      exp = "#{column_sql} AS #{field_name}"
      assert_equal exp, expression.args.first
    end

    should "know how to add a select for a field with a column SQL block" do
      field_name = Factory.string
      MR::ReadModel.add_select_for_field(@read_model_class, field_name){ |sql| sql }

      assert_equal 1, @read_model_class.relation.query_expressions.size
      expression = @read_model_class.relation.query_expressions.first
      assert_equal :select, expression.type

      column_sql = Factory.string
      exp = "#{column_sql} AS #{field_name}"
      assert_equal exp, expression.block.call(column_sql)
    end

  end

  class MixinTests < UnitTests
    desc "mixed in"
    setup do
      @struct_class = Class.new{ include MR::ReadModelStruct }

      @add_select_for_field_called_with = nil
      Assert.stub(MR::ReadModel, :add_select_for_field) do |*args, &block|
        @add_select_for_field_called_with = args + [block]
        Assert.stub_send(MR::ReadModel, :add_select_for_field, *args, &block)
      end
    end
    subject{ @read_model_class }

    should have_imeths :field, :json_struct_list, :json_struct_obj

    should "use the mr read model struct mixin" do
      assert_includes MR::ReadModelStruct, subject
    end

    should "use the mr read model querying mixin" do
      assert_includes MR::ReadModel::Querying, subject
    end

    should "know how to add fields with a select" do
      field_name       = Factory.string
      column_sql       = Factory.string
      column_sql_block = proc{ |sql| sql }
      subject.field(field_name, :string, column_sql, &column_sql_block)

      field = subject.fields.find(field_name)
      assert_instance_of MR::ReadModel::Field, field
      assert_equal :string, field.type
      assert_respond_to field.name, subject.new

      exp = [subject, field_name, column_sql, column_sql_block]
      assert_equal exp, @add_select_for_field_called_with
    end

    should "know how to add json struct lists with a select" do
      field_name       = Factory.string
      column_sql       = Factory.string
      column_sql_block = proc{ |sql| sql }
      subject.json_struct_list(field_name, @struct_class, column_sql, &column_sql_block)

      field = subject.json_struct_lists.find{ |f| f.name == field_name }
      assert_instance_of MR::ReadModel::JsonStructListField, field
      assert_equal @struct_class, field.struct_class
      assert_respond_to field.name, subject.new

      exp = [subject, field_name, column_sql, column_sql_block]
      assert_equal exp, @add_select_for_field_called_with
    end

    should "know how to add json struct objs with a select" do
      field_name       = Factory.string
      column_sql       = Factory.string
      column_sql_block = proc{ |sql| sql }
      subject.json_struct_obj(field_name, @struct_class, column_sql, &column_sql_block)

      field = subject.json_struct_objs.find{ |f| f.name == field_name }
      assert_instance_of MR::ReadModel::JsonStructObjField, field
      assert_equal @struct_class, field.struct_class
      assert_respond_to field.name, subject.new

      exp = [subject, field_name, column_sql, column_sql_block]
      assert_equal exp, @add_select_for_field_called_with
    end

  end

  class ReadModelStructTests < Assert::Context
    desc "MR::ReadModelStruct"
    setup do
      @read_model_struct_class = Class.new do
        include MR::ReadModelStruct
      end
    end
    subject{ @read_model_struct_class }

    should "use much-plugin" do
      assert_includes MuchPlugin, MR::ReadModelStruct
    end

    should "use the mr read model data mixin" do
      assert_includes MR::ReadModel::Data, subject
    end

    should "use the mr read model fields mixin" do
      assert_includes MR::ReadModel::Fields, subject
    end

  end

  class ReadModelStructInitTests < ReadModelStructTests
    desc "when init"
    setup do
      @read_model_struct_class.class_eval do
        field :name, :string
      end

      @data       = { 'name' => Factory.string }
      @read_model = @read_model_struct_class.new(@data)
    end
    subject{ @read_model }

    should "set its data" do
      assert_equal @data['name'], subject.name
    end

    should "not error when initialized without data" do
      read_model_struct = @read_model_struct_class.new
      assert_nil read_model_struct.name
    end

    should "be comparable" do
      matching_struct = @read_model_struct_class.new(@data)
      assert_equal matching_struct, subject

      data = Factory.boolean ? { 'name' => Factory.string } : {}
      not_matching_struct = @read_model_struct_class.new(data)
      assert_not_equal not_matching_struct, subject
    end

    should "know its inspect" do
      object_hex     = (subject.object_id << 1).to_s(16)
      values_inspect = @read_model_struct_class.fields.map do |field|
        "#{field.ivar_name}=#{field.read(@data).inspect}"
      end.join(' ')
      exp = "#<#{@read_model_struct_class}:0x#{object_hex} #{values_inspect}>"
      assert_equal exp, subject.inspect
    end

  end

end
